static char rcsid[] = "@(#)$Id: curs_input.c,v 1.2 2003/10/01 17:34:34 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.2 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************
 *
 * Some code copied from ../src/curses.c. It have following copyright:
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/


#include "def_screen.h"

DEBUG_VAR(Debug,__FILE__,"screen");

#include    <errno.h>


#if POLL_METHOD
static unsigned char *read_buffer   = NULL;
static int read_buffer_len = 0;
static int last_read_error = 0;

static int background_read P_((int fd, void *data));
static int background_read(fd,data)
     int fd; 
     void *data;
{
    int    result;
    unsigned char   ch;
    
    result = read(fd, &ch, 1);

    if (result < 0) {
	int err = errno;

	SIGDPRINT(Debug,4,(&Debug, 
			   "background_read: errno = %d [%s]\n",err,
			   error_description(err)));

	if((err == EINTR)
#ifdef	EAGAIN
	   || (err == EAGAIN)
#endif
#ifdef	EWOULDBLOCK
	   || (err == EWOULDBLOCK)
#endif
	   ) {
	    return 1;  /* Retry */
	}
	
	last_read_error = err;
	return 0;
    }
    
    if (0 == result) {
	SIGDPRINT(Debug,4,(&Debug,
			   "background_read: Got zero bytes...\n"));
	last_read_error = -1;
	return 0;
    }

    read_buffer = safe_realloc(read_buffer,read_buffer_len+1);
    SIGDPRINT(Debug,20,(&Debug,
	       "background_read(fd=%d)=1 [%d]",fd,read_buffer_len));
    SIGDPRINT(Debug,50,(&Debug,"=%d",ch));
    SIGDPRINT(Debug,20,(&Debug,"\n"));
    
    read_buffer[read_buffer_len++] = ch; 
    last_read_error = 0;

    return 1;
}


void clear_input_buffer() 
{
    SIGDPRINT(Debug,4,(&Debug, 
		       "Resetting background_read buffer (%d bytes)...\n",
		       read_buffer_len));
    read_buffer_len = 0;
    last_read_error = 0;
}

int error_sleep(seconds) 
     int seconds;
{
    int ret;
    if (read_buffer_len > 0) {
	SIGDPRINT(Debug,5,(&Debug, 
			   "error_sleep(%d)=1  -- sleep skipped\n",
			   seconds));
	return 1;
    }
    SIGDPRINT(Debug,5,(&Debug, 
		       "error_sleep(%d) ...\n",seconds));
    ret = wait_for_action_or_timeout(background_read,seconds);
    DPRINT(Debug,5,(&Debug, 
		    "error_sleep=%d\n",ret));
    return ret;
}
#endif


int cur_ReadCh(flags,break_flag)
     int flags;
     int break_flag;
{
    int ret = 0;
    
    do {
	struct charset_state *ch = cur_ReadCh2(flags,break_flag);

	if (!ch) {
	    ret = EOF;
	    DPRINT(Debug,4,(&Debug, 
			    "ReadCh=EOF (errno=%d)\n",errno));
	} else if (ch -> caller_flags) {
	    ret = ch -> caller_flags;
	    DPRINT(Debug,4,(&Debug, 
			    "ReadCh=%d (flags)\n",ret));
	} else
	    ret =  state_is_onebyte(ch);
	if (!ret) {
	    DPRINT(Debug,4,(&Debug,
			    "ReadCh: Not a one byte character or zero byte!"));
	    /* Ring a bell */
	    Writechar('\007');
	} else {
	    DPRINT(Debug,4,(&Debug,
			    "ReadCh=%d (one byte character)\n",
			    ret));
	}
    } while (0 == ret);
    return ret;
}


struct charset_state * cur_ReadCh2(flags,break_flag)
     int flags;
     int break_flag;
{
    int redraw      = (flags & READCH_MASK);
    int cursorkeys  = (flags & READCH_CURSOR)    != 0;
    int nocursor    = (flags & READCH_NOCURSOR)  != 0 && cursor_control;
    int term_char   = (flags & READCH_term_char) != 0;
    int x,y;

    unsigned char input_buffer[20];

    /*
     *	read a character with Raw mode set!
     *
     *	EAGAIN & EWOULDBLOCK are recognized just in case
     *	O_NONBLOCK happens to be in effect.
     */

    /* This is static array so we can initialize it in here ...
     */
    static struct {
	char ** CONST str;
	CONST int result;
	int maybe;
    } keytable[] =
      { { &_key_up,       UP_MARK,       0 },
	{ &_key_down,     DOWN_MARK,     0 },
	{ &_key_left,     LEFT_MARK,     0 },
	{ &_key_right,    RIGHT_MARK,    0 },

	{ &_key_pageup,   PAGEUP_MARK,   0 },
	{ &_key_pagedown, PAGEDOWN_MARK, 0 },

	{ &_key_home,     HOME_MARK,     0 },
	{ &_key_help,     HELP_MARK,     0 },
	{ &_key_find,     FIND_MARK,     0 },
	{ NULL,           0,             0 }
      };
    
    int read_p,found_key;

#if POLL_METHOD
    change_action(terminal_fd,0,background_read,no_action_routine,
		  no_action_routine,
		  NULL);
#endif    

    cur_GetXYLocation(&x,&y); /* Background actions may print error messages .. */

    if (!last_state) {
	last_state = new_state(system_charset);
	DPRINT(Debug,4,(&Debug,
		       "Keyboard input charset is %s\n",
		       last_state->charset->MIME_name ?
		       last_state->charset->MIME_name :
		       "<no MIME name>"));
    }

    if (state_ready(last_state)) {
	DPRINT(Debug,4,(&Debug, 
			"ReadCh2: starting reading new character (set %s)\n",
			last_state->charset->MIME_name ? 
			last_state->charset->MIME_name :
			"<no MIME name>"));
	reset_state(last_state,0);
    }
    last_state->caller_flags = 0;

 reinit_ReadChar:
    read_p = 0;
    found_key = 0;
    
    FlushBuffer();

    if (redraw && !cur_RawState()) { /* Check that we have in 'raw' mode */
	DPRINT(Debug,4,(&Debug,
			"ReadCh2: Going to Raw mode\n"));
	cur_Raw(ON);
	cur_ClearScreen();

	reset_state(last_state,1);
	last_state->caller_flags = redraw;

	return last_state;
    }

    if (redraw && redraw_screen) {
	DPRINT(Debug,4,(&Debug,
			"ReadCh2: Pending redraw...\n"));
	cur_ClearScreen();

	reset_state(last_state,1);
	last_state->caller_flags = redraw;

	return last_state;
    }
  
    if ((_intransmit != ON || redraw_screen) &&
	cursorkeys && _transmit_on) {
	DPRINT(Debug,4,(&Debug,
			"ReadCh2: Enabling cursor keys...\n"));
	transmit_functions(ON);
    }

    if ((_intransmit != OFF || redraw_screen) &&
	nocursor && _transmit_off) {
	DPRINT(Debug,4,(&Debug,
			"ReadCh2: Disabling cursor keys...\n"));
	transmit_functions(OFF);
    }
    
    if (cursorkeys) {
	int i;
	DPRINT(Debug,8,(&Debug,
			"ReadCh2: Available function keys:"));

	for (i = 0; keytable[i].str != 0; i++) {
	    char * CONST str = *(keytable[i].str);
	    if(str && str[0] != '\0') {
		keytable[i].maybe = 1;  /* Initially every function key is possible */
		SIGDPRINT(Debug,8,(&Debug,
				   " [%d] %d",i,keytable[i].result));	     
	    }
	}      
	DPRINT(Debug,4,(&Debug,
			"\n"));
    }

    while (found_key == 0) {
	unsigned char   ch;
#if POLL_METHOD
	if (read_buffer_len < 1 && last_read_error == 0) {
	    errno= 0;
	    if(!wait_for_action(background_read)) {
		if (errno != 0) {
		    DPRINT(Debug,4,(&Debug,
				    "ReadCh2: wait_for_action got error %d (%s)\n",
				    errno, error_description(errno)));
		    /* Return error: */
		    if ((redraw && redraw_screen)
			|| break_flag   /* GetPrompt wants to see errors! */
			) {
			DPRINT(Debug,4,(&Debug,
					"       ... return error to caller\n"));
			found_key = -1;
			continue;
		    }
		    if (errno == EINTR
#ifdef	EAGAIN
			|| (errno == EAGAIN)
#endif
#ifdef	EWOULDBLOCK
			|| (errno == EWOULDBLOCK)
#endif
			) {
			DPRINT(Debug,4,(&Debug,
					"       ... retry (ignored)\n"));
		    } else
			found_key = -1;
		    continue;		    
		}
		DPRINT(Debug,4,(&Debug,
				"ReadCh2: interrupted (ignored)\n"));
		continue;
	    }
	}
	if (read_buffer_len > 0) {
	    ch = read_buffer[0];
	    if (--read_buffer_len > 0)
		memmove(read_buffer,read_buffer+1,read_buffer_len);
	} else if (last_read_error != 0) {
	    if (-1 == last_read_error) {
		DPRINT(Debug,4,(&Debug,
				"ReadCh2: background_read got zero bytes...\n"));
		found_key = -1;
		continue;
	    }
	    /* Return error: */
	    if ((redraw && redraw_screen)
		|| break_flag   /* GetPrompt wants to see errors! */
		) {
		found_key = -1;
		continue;
	    }
	    if (last_read_error != 0) {
		errno = last_read_error;
		DPRINT(Debug,4,(&Debug,
				"ReadCh2: background_read got error %d (%s)\n",
				last_read_error,
				error_description(last_read_error)));
		found_key = -1;
	    }
	    continue;         /* continue or quit */	    
	} else {
	    DPRINT(Debug,1,(&Debug,
			    "ReadCh2: SOFTWARE ERROR !!\n"));
	    break;
	}
#else
	int    result;

	result = read(terminal_fd, &ch, 1);
	
	if (result < 0) {
	    int err = errno;
	    DPRINT(Debug,4,(&Debug,
			    "ReadCh2: errno = %d [%s]\n",err,
			    error_description(err)));
	    
	     /* Return error: */
	    if (redraw && redraw_screen
		|| break_flag   /* GetPrompt wants to see errors! */
		) {
		found_key = -1;
		continue;
	    }
	    if((errno == EINTR)
#ifdef	EAGAIN
	       || (errno == EAGAIN)
#endif
#ifdef	EWOULDBLOCK
	       || (errno == EWOULDBLOCK)
#endif
	       ) {
		continue;  /* Retry */
	    }
	    break;
	}
	if (0 == result) {
	    DPRINT(Debug,4,(&Debug,
			    "ReadCh2: Got zero bytes...\n"));
	    found_key = -1;
	    continue;
	}
#endif

#ifdef USE_DLOPEN
	{
	    union xxx_rand {
		int ch;
		char bytes[sizeof (int)];
	    } A;
	    A.ch = ch;

	    seed_rand_bits(A.bytes, sizeof A, 
			   3 /* Assume 3 bits per character ... */);
	}
#endif

	

	DPRINT(Debug,50,(&Debug,
			 "ReadCh2: Looking char %d (read_p=%d)\n",ch,read_p));

	if (term_char) {
	    if (backspace == ch) 
		found_key = TERMCH_backspace;
	    if (kill_line == ch)
		found_key = TERMCH_kill_line;
	    if (word_erase  == ch)
		found_key = TERMCH_word_erase;
	    if (interrupt_char  == ch)
		found_key = TERMCH_interrupt_char;
	    if (reprint_char    == ch)
		found_key = TERMCH_reprint_char;
	    if (eof_char        == ch)
		found_key = TERMCH_eof_char;
	    if (found_key) {
		DPRINT(Debug,4,(&Debug,
				"ReadCh2: found termchar = %d (char=%02X)\n",
				found_key,ch));
		/* Soft reset state */
		reset_state(last_state,0); 
		break;
	    }
	}

	if (cursorkeys) {
	    int match = 0;
	    int i;
	    for (i = 0; keytable[i].str != NULL; i++) {
		if (keytable[i].maybe) {
		    unsigned char * CONST str = 
			(unsigned char *) *(keytable[i].str);
		    if (str[read_p] == ch) {
			match++;
			if (str[read_p+1] == '\0') {
			    found_key = keytable[i].result;
			    DPRINT(Debug,4,(&Debug,
					    "ReadCh2: Found function key = %d (keytable = %d, read_p =%d)\n",
					    found_key,i,read_p));

			    /* Soft reset state */
			    reset_state(last_state,0); 
			}
		    } else {
			keytable[i].maybe = 0;
		    }
		}
	    }

	    if (read_p < sizeof input_buffer -1) {
		input_buffer[read_p] = ch;
		input_buffer[read_p+1] = '\0';
	    }

	    if (match == 0) {    /* Not in keytable */
		if (read_p == 0) 
		    found_key = ch;  /* Normal key */
		else {
		    int i;
		    
		    /* But maybe escape sequence is valid state change ? */
		    	
		    for (i = 0; 
			 i <= read_p && i < sizeof input_buffer -1; 
			 i++) {

			if (state_ready(last_state)) {
			    DPRINT(Debug,4,(&Debug,
					    "ReadCh2: Unsupporting -- function keys and state change sequences overlap? (OR BAD SEQUENCE) \n"));
			    goto BAD_sequence;
			}
			
			if (!add_streambyte_to_state(last_state,
						     input_buffer[i]))
			    goto BAD_sequence;
			
		    }
		    
		    if (state_ready(last_state))
			goto got_key;
		    DPRINT(Debug,4,(&Debug,
				    "ReadCh2: need more bytes for character...\n"));
		    goto reinit_ReadChar;	    


		BAD_sequence:

		    /* Soft reset state */
		    reset_state(last_state,0); 

		    DPRINT(Debug,4,(&Debug,
				    "ReadCh2: Bad escape sequence; ch = %d, read_p = %d\n",
				    ch,read_p));
#ifdef DEBUG
		    DPRINT(Debug,4,(&Debug,
				    "ReadCh2: Sequence was:"));
		    for (i = 0; i <= read_p && i < sizeof input_buffer -1; i++) {
			if (isascii(input_buffer[i]) && isprint(input_buffer[i])) {
			    DPRINT(Debug,4,(&Debug," %c (0x%02X)", 
					    input_buffer[i],input_buffer[i]));
			} else {
			    DPRINT(Debug,4,(&Debug," 0x%02X", input_buffer[i]));
			}
		    }
		    if (read_p > sizeof input_buffer -1)
			DPRINT(Debug,4,(&Debug," ..."));
		    DPRINT(Debug,4,(&Debug,"\n"));
#endif

		    /* Ring a bell */
		    cur_Writechar('\007');
		    goto reinit_ReadChar;
		}
	    } else
		read_p++;
	} else
	    found_key = ch;
	
    }

    if (found_key <= 0 && redraw && redraw_screen) {
	DPRINT(Debug,4,(&Debug,
			"ReadCh2: Redraw...\n"));
	if(!cur_RawState()) {  /* Check that we have in 'raw' mode */
	    DPRINT(Debug,4,(&Debug,
			    "ReadCh2: Going to Raw mode\n"));
	    cur_Raw(ON);      
	}
	cur_ClearScreen();

	reset_state(last_state,1);
	last_state->caller_flags = redraw;

	return last_state;
    }

    cur_MoveCursor(x,y);  
    DPRINT(Debug,50,(&Debug,
		     "ReadCh2: found_key=%d\n",found_key));
    
    if (found_key < 0) {
	reset_state(last_state,1);
	DPRINT(Debug,4,(&Debug,
			"ReadCh2=NULL (state)\n")); 
	return NULL;
    } else {
	if (found_key >= 256) {
	    reset_state(last_state,1);
	    last_state->caller_flags = found_key;
	} else {
	    last_state->caller_flags = 0;
	    if (!add_streambyte_to_state(last_state,found_key)) {
		/* Ring a bell */
		Writechar('\007');
		DPRINT(Debug,4,(&Debug,
				"ReadCh2: bad sequence...\n"));
		reset_state(last_state,1);
	    }
	}
    }
    
    if (!state_ready(last_state) &&
	!last_state->caller_flags) {
	DPRINT(Debug,4,(&Debug,
			"ReadCh2: need more bytes for character...\n"));
	goto reinit_ReadChar;
    }

 got_key:
    DPRINT(Debug,4,(&Debug,
		    "ReadCh2=%p (state): caller_flags=%d, ready=%d\n",
		    last_state,last_state->caller_flags,
		    state_ready(last_state))); 
    return last_state;
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
