static char rcsid[] = "@(#)$Id: screen.c,v 1.4 2003/10/12 06:42:38 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.4 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 * Some code copied from ../curses.c. It have following copyright:
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *
 * However on that code there is no practically Elm 2.4 code (from where that
 * copyright is) left. Only function names are same.
 *****************************************************************************/



#include "def_screen.h"

DEBUG_VAR(Debug,__FILE__,"screen");

static void PutLineS P_((struct string *S));

static void PutLineS(S)
     struct string *S;
{

    DPRINT(Debug,49, (&Debug, "PutLineS(\"%S\")\n",S));

    cur_PutLineS(S);
}


int InitScreen()
{
    int r;

    DPRINT(Debug,49, (&Debug, "InitScreen()\n"));

    r = cur_InitScreen();

    set_start_run_hooks(print_status,RawState,Raw,
			start_run_tty_init,ClearScreen,
			PutLineS, print_status_cooked
			);

    return r;
}


/* This assumes one byte charset */
void Writechar(ch)
     int ch;
{
    DPRINT(Debug,49, (&Debug, "Writechar('%c')\n",ch));

    cur_Writechar(ch);
}

/*VARARGS2*/

void Write_to_screen (
#if ANSI_C
		      CONST char *format, CONST char *msg, ...
#else
		      format, msg, va_alist
#endif
		      )
#if !ANSI_C
     CONST char *format; 
     CONST char *msg;
     va_dcl
#endif
{
    va_list vl;
    struct string *text;

    DPRINT(Debug,49, (&Debug, "Write_to_screen(format=\"%s\",...)\n",format));

    Va_start(vl, msg);           /* defined in defs.h */
    
    /** This routine writes to the screen at the current location.
	when done, it increments lines & columns accordingly by
	looking for "\n" sequences... **/

    text = elm_smessage(0,format,msg,vl);
    PutLineS(text);
    free_string(&text);
    
    va_end(vl);
    
}

void PutLine0(x, y, line)
     int x,y;
     CONST char *line;
{
    DPRINT(Debug,49, (&Debug, "PutLine0(%d,%d,\"%s\")\n",x,y,line));

    cur_PutLine0(x,y,line);
}


void PutLineX(
#if ANSI_C
     int x, int y, const char *format, const char *line, ...
#else
     x, y, format, line, va_alist
#endif
     )
#if !ANSI_C
     int x; 
     int y; 
     CONST char *format; 
     CONST char *line; 
     va_dcl
#endif
{
    va_list vl;
    struct string *text;

    DPRINT(Debug,49, (&Debug, "PutLineX(%d,%d,format=\"%s\",...)\n",x,y,format));

    Va_start(vl, line);           /* defined in defs.h */
    
    MoveCursor(x,y);
    
    text = elm_smessage(0,format,line,vl);
    PutLineS(text);
    free_string(&text);

    va_end(vl);
    
}


int CleartoEOLN()
{
    int r;

    DPRINT(Debug,49, (&Debug, "CleartoEOLN()\n"));
    r = cur_CleartoEOLN();

    return r;
}

int CleartoEOS()
{
    int r;

    DPRINT(Debug,49, (&Debug, "CleartoEOS()\n"));
    r = cur_CleartoEOS();

    return r;
}

int RawState()
{
    int r;

    DPRINT(Debug,49, (&Debug, "RawState()\n"));

    r = cur_RawState();

    return r;
}

void Raw(state)
     int state;
{
    DPRINT(Debug,49, (&Debug, "Raw(%d)\n",state));

    cur_Raw(state);
}

struct charset_state * ReadCh2(flags)
     int flags;
{
    struct charset_state *st;

    DPRINT(Debug,49, (&Debug, "ReadCh2(%d)\n",flags));

    st = cur_ReadCh2(flags,InGetPrompt);

    return st;
}

int ReadCh(flags)
     int flags;
{
    int c;

    DPRINT(Debug,49, (&Debug, "ReadCh(%d)\n",flags));

    c = cur_ReadCh(flags,InGetPrompt);

    return c;
}

void ScreenSize(lines, columns)
     int *lines, *columns;
{

    DPRINT(Debug,49, (&Debug, "ScreenSize(*,*)\n"));

    /* cur_ScreenSize   DO NOT return lines-1 (as ScreenSize() did) */

    cur_ScreenSize(lines,columns);

    ((*lines)--);

}

void GetXYLocation(x,y)
     int *x,*y;
{
    DPRINT(Debug,49, (&Debug, "GetXYLocation(*,*)\n"));

    cur_GetXYLocation(x,y);
}

int ClearScreen()
{
    int r;

    DPRINT(Debug,49, (&Debug, "ClearScreen()\n"));

    r = cur_ClearScreen();

    return r;
}

int MoveCursor(row, col)
     int row, col;
{
    int r;

    DPRINT(Debug,49, (&Debug, "MoveCursor(%d,%d)\n",row,col));

    r = cur_MoveCursor(row, col);

    return r;
}

void CarriageReturn()
{

    DPRINT(Debug,49, (&Debug, "CarriageReturn()\n"));
    /** move the cursor to the beginning of the current line **/
    Writechar('\r');
}

void NewLine()
{
    DPRINT(Debug,49, (&Debug, "NewLine()\n"));
    /** move the cursor to the beginning of the next line **/

    Writechar('\r');
    Writechar('\n');
    FlushBuffer();	
}

int modes = 0;

int StartBold()
{
    DPRINT(Debug,49, (&Debug, "StartBold()\n"));
    cur_changemode(&modes, pg_BOLD,0);

    return 0;
}

int EndBold()
{
    DPRINT(Debug,49, (&Debug, "EndBold()\n"));
    /** compliment of startbold **/

    cur_changemode(&modes, 0, pg_BOLD);

    return 0;
}

int StartInverse() {

    DPRINT(Debug,49, (&Debug, "StartInverse()\n"));
    cur_changemode(&modes, pg_INVERSE,0);
    
    return 0;
}

int EndInverse() {

    DPRINT(Debug,49, (&Debug, "EndInverse()\n"));
    cur_changemode(&modes, 0, pg_INVERSE);
    
    return 0;
}


int StartUnderline() {

    DPRINT(Debug,49, (&Debug, "StartUnderline()\n"));
    cur_changemode(&modes, pg_UNDERLINE,0);
    
    return 0;
}

int EndUnderline() {

    DPRINT(Debug,49, (&Debug, "EndUnderline()\n"));
    cur_changemode(&modes, 0, pg_UNDERLINE);
    
    return 0;
}

int StartHalfbright() {

    DPRINT(Debug,49, (&Debug, "StartHalfbright()\n"));
    cur_changemode(&modes, pg_HALFBRIGHT,0);
    
    return 0;
} 

int EndHalfbright() {

    DPRINT(Debug,49, (&Debug, "EndHalfbright()\n"));
    cur_changemode(&modes, 0, pg_HALFBRIGHT);

    return 0;
}

int get_tabspacing()
{
    DPRINT(Debug,49, (&Debug, "get_tabspacing()\n"));
    return cur_tabspacing;
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
